/**
 * FLL - Level 0
 *
 * Project: Memory
 * API Version: 0.7
 * Licenses: lgpl-2.1-or-later
 *
 * Provide means to use memory routines, with error checking.
 */
#ifndef _F_memory_h
#define _F_memory_h

// Libc includes.
#include <malloc.h>
#include <stdlib.h>
#include <string.h>

// FLL-0 includes.
#include <fll/level_0/type.h>
#include <fll/level_0/status.h>

// FLL-0 memory includes.
#include <fll/level_0/memory/common.h>
#include <fll/level_0/memory/array.h>
#include <fll/level_0/memory/arrays.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Securely resize dynamically allocated data.
 *
 * Will change all data to 0 prior to deallocation.
 *
 * The pointer address might be changed by realloc().
 *
 * @param length_old
 *   The total number of blocks representing the length to be resized from.
 * @param length_new
 *   The total number of blocks representing the length to be resized to.
 * @param size
 *   The block size, in bytes (size * length = allocated size).
 * @param pointer
 *   A pointer to the address that will be resized.
 *
 * @return
 *   F_okay on success.
 *   F_data_not on success, but there was nothing to do.
 *
 *   F_memory_not (with error bit) on out of memory.
 *   F_parameter (with error bit) if a parameter is invalid.
 *
 * @see calloc()
 * @see memset()
 * @see realloc()
 */
#ifndef _di_f_memory_adjust_
  extern f_status_t f_memory_adjust(const size_t length_old, const size_t length_new, const size_t size, void ** restrict const pointer);
#endif // _di_f_memory_adjust_

/**
 * Delete dynamically allocated data.
 *
 * Will not change any of the data to 0 prior to deallocation.
 *
 * Type and length are not normally used by this function but must be provided for the cases when f_memory_delete is swapped with f_memory_destroy (or vice-versa).
 *
 * @param length
 *   The total number of blocks to be allocated.
 *   If length is 0 then no delete is performed.
 * @param size
 *   The block size, in bytes (size * length = allocated size).
 *   If size is 0 then no delete is performed.
 * @param pointer
 *   A pointer to the address that will be freed.
 *
 * @return
 *   F_okay on success.
 *   F_data_not on success, but there was nothing to do.
 *
 *   F_parameter (with error bit) if a parameter is invalid.
 *
 * @see free()
 */
#ifndef _di_f_memory_delete_
  extern f_status_t f_memory_delete(const size_t length, const size_t size, void ** restrict const pointer);
#endif // _di_f_memory_delete_

/**
 * Securely deletes some dynamically allocated data.
 *
 * Will change all data to 0 prior to deallocation.
 *
 * @param length
 *   The total number of blocks to be allocated.
 *   If length is 0 then no delete is performed.
 * @param size
 *   The block size, in bytes (size * length = allocated size).
 *   If size is 0 then no delete is performed.
 * @param pointer
 *   A pointer to the address that will be freed.
 *
 * @return
 *   F_okay on success.
 *   F_data_not on success, but there was nothing to do.
 *
 *   F_parameter (with error bit) if a parameter is invalid.
 *
 * @see free()
 * @see memset()
 */
#ifndef _di_f_memory_destroy_
  extern f_status_t f_memory_destroy(const size_t length, const size_t size, void ** restrict const pointer);
#endif // _di_f_memory_destroy_

/**
 * Create some dynamically allocated array of some length.
 *
 * @param length
 *   The total number of blocks to be allocated.
 *   Must be greater than 0.
 * @param pointer
 *   A pointer that will be updated to the address of the newly allocated memory.
 * @param size
 *   The block size, in bytes (size * length = allocated size).
 *   Must be greater than 0.
 *
 * @return
 *   F_okay on success.
 *   F_data_not on success, but there was nothing to do.
 *
 *   F_memory_not (with error bit) on allocation error.
 *   F_parameter (with error bit) if a parameter is invalid.
 *
 * @see calloc()
 * @see memset()
 */
#ifndef _di_f_memory_new_
  extern f_status_t f_memory_new(const size_t length, const size_t size, void ** restrict const pointer);
#endif // _di_f_memory_new_

/**
 * Create some dynamically allocated array of some length, guaranteeing aligned memory.
 *
 * @param length
 *   The total number of blocks to be allocated.
 *   Must be greater than 0.
 *   Must be a multiple of alignment.
 * @param alignment
 *   The size of the alignment, such as sizeof(void *).
 *   Must be greater than 0.
 * @param pointer
 *   A pointer that will be updated to the address of the newly allocated memory.
 *
 * @return
 *   F_okay on success.
 *   F_data_not on success, but there was nothing to do.
 *
 *   F_memory_not (with error bit) on allocation error.
 *   F_parameter (with error bit) if a parameter is invalid.
 *
 * @see aligned_alloc()
 * @see posix_memalign()
 * @see memset()
 */
#ifndef _di_f_memory_new_aligned_
  extern f_status_t f_memory_new_aligned(const size_t length, const size_t alignment, void ** restrict const pointer);
#endif // _di_f_memory_new_aligned_

/**
 * Resize dynamically allocated data.
 *
 * Will not change any of the data prior to deallocation.
 *
 * The pointer address might be changed by realloc().
 *
 * @param length_old
 *   The total number of blocks representing the length to be resized from.
 * @param length_new
 *   The total number of blocks representing the length to be resized to.
 * @param size
 *   The block size, in bytes (size * length = allocated size).
 * @param pointer
 *   A pointer to the address that will be resized.
 *
 * @return
 *   F_okay on success.
 *   F_data_not on success, but there was nothing to do.
 *
 *   F_memory_not (with error bit) on out of memory.
 *   F_parameter (with error bit) if a parameter is invalid.
 *
 * @see calloc()
 * @see memset()
 * @see realloc()
 */
#ifndef _di_f_memory_resize_
  extern f_status_t f_memory_resize(const size_t length_old, const size_t length_new, const size_t size, void ** restrict const pointer);
#endif // _di_f_memory_resize_

#ifdef __cplusplus
} // extern "C"
#endif

#endif // _F_memory_h
